"""Agents Assembly translator"""

__version__ = "0.1.5"

from aasm.generating.code import Code
from aasm.modules.module import Module
from aasm.generating.python_spade import get_spade_code
from aasm.generating.python_module import get_modules_for_target
from aasm.utils.exception import PanicException
